import numpy as np


class UpliftModel:
    """Class of Uplift Modeling."""

    def __init__(self, learner_treat, learner_control):
        """

        Parameters
        ----------
        learner_treat :
            Learner to estimate effect of treatment group.
        learner_control :
            Learner to estimate effect of control group.
        """
        self.learner_treat = learner_treat
        self.learner_control = learner_control

    def fit(
        self,
        X_treat,
        y_treat,
        X_control,
        y_control,
        weight_treat=None,
        weight_control=None,
    ):
        """

        Parameters
        ----------
        X_treat : numpy.ndarray
            Features for learner_treat.
        y_treat : numpy.ndarray
            Labels for learner_treat.
        X_control : numpy.ndarray
            Features for learner_control.
        y_control : numpy.ndarray
            Labels for learner_control.
        weight_treat : numpy.ndarray or None
            Weights for learner_treat.
        weight_control : numpy.ndarray or None
            Weights for learner_control.

        Returns
        -------
        None
        """
        self.learner_treat.fit(X_treat, y_treat, sample_weight=weight_treat)
        self.learner_control.fit(
            X_control, y_control, sample_weight=weight_control
        )

    def estimate_uplift_score(self, X):
        """Estimate uplift scores.

        Parameters
        ----------
        X : numpy.ndarray
            Features for prediction treat and control probability.

        Returns
        -------
        uplift_score : np.array
            Uplift Score.
        """
        proba_treat = self.learner_treat.predict_proba(X)[:, 1]
        proba_control = self.learner_control.predict_proba(X)[:, 1]
        uplift_score = proba_treat / proba_control
        return uplift_score

    def predict(self, X, treatment, y):
        """

        Parameters
        ----------
        X : numpy.ndarray
            Features for prediction treat and control probability.
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.

        Returns
        -------
        (uplift_score, lift) : tuple
            Uplift score and lift values.
        """
        uplift_score = self.estimate_uplift_score(X)

        sorted_idx = np.argsort(uplift_score)[::-1]
        uplift_score = uplift_score[sorted_idx]
        y = y[sorted_idx]
        treatment = treatment[sorted_idx]

        y_treat = np.nancumsum(np.where(treatment == 1, y, np.nan))
        y_control = np.nancumsum(np.where(treatment == 0, y, np.nan))

        treat_size = np.nancumsum(np.where(treatment == 1, treatment, np.nan))
        control_size = np.nancumsum(
            np.where(treatment == 0, (1 - treatment), np.nan)
        )

        cumavg_y_treat = np.array(
            [0.0 if s == 0 else _y / s for _y, s in zip(y_treat, treat_size)]
        )
        cumavg_y_control = np.array(
            [
                0.0 if s == 0 else _y / s
                for _y, s in zip(y_control, control_size)
            ]
        )

        lift = (cumavg_y_treat - cumavg_y_control) * treat_size
        return (uplift_score, lift)

    def get_baseline(self, lift):
        """

        Parameters
        ----------
        lift : numpy.ndarray
            Array of lift, treatment effect.

        Returns
        -------
        base_line : numpy.ndarray
            Array of random treat effect.
        """
        data_size = len(lift)
        base_line = np.arange(data_size) * lift[data_size - 1] / data_size
        return base_line

    def get_auuc(self, lift):
        """

        Parameters
        ----------
        lift : numpy.ndarray
            Array of lift, treatment effect.

        Returns
        -------
        auuc : float
            AUUC score.
        """
        base_line = self.get_baseline(lift)
        auuc = (lift - base_line).sum() / len(lift)
        return auuc



from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn import metrics
from statsmodels.regression import linear_model


class EffectSize:
    """Calculating the effect size-d.

    Examples
    --------
    ate_weight = model.get_weight(treatment, mode='ate')
    es = metrics.EffectSize()
    es.fit(X, treatment, weight=ate_weight)
    es.transform() # return (effect_size, effect_name)
    """

    def __init__(self) -> None:
        self.effect_size = None
        self.effect_name = None

    def fit(
        self,
        X: pd.DataFrame,
        treatment: np.ndarray,
        weight: Union[np.ndarray, None] = None,
    ) -> None:
        """Fit the model with X.

        Parameters
        ----------
        X : pd.DataFrame
            Covariates for propensity score.
        treatment : pd.Series
            Flags with or without intervention.
        weight : np.array
            The weight of each sample.

        Returns
        -------
        None
        """
        if weight is None:
            weight = np.ones(X.shape[0])
        # Cauculation Average and Variance of Treat Group.
        treat_avg = np.average(X[treatment], weights=weight[treatment], axis=0)
        treat_var = np.average(
            np.square(X[treatment] - treat_avg),
            weights=weight[treatment],
            axis=0,
        )
        # Cauculation Average and Variance of Control Group.
        control_avg = np.average(
            X[~treatment], weights=weight[~treatment], axis=0
        )
        control_var = np.average(
            np.square(X[~treatment] - control_avg),
            weights=weight[~treatment],
            axis=0,
        )
        # Estimate d_value.
        data_size = X.shape[0]
        treat_size = np.sum(treatment)
        control_size = np.sum(~treatment)
        sc = np.sqrt(
            (treat_size * treat_var + control_size * control_var) / data_size
        )
        d_value = np.abs(treat_avg - control_avg) / sc

        self.effect_size = d_value
        self.effect_name = X.columns.to_numpy()

    def transform(self):
        """Apply the calculating the effect size d.

        Returns
        -------
        (effect_name, effect_size) : tuple
        """
        return (self.effect_name, self.effect_size)

    def fit_transform(
        self,
        X: pd.DataFrame,
        treatment: np.ndarray,
        weight: Union[np.ndarray, None] = None,
    ):
        """Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : pd.DataFrame
            Covariates for propensity score.
        treatment : pd.Series
            Flags with or without intervention.
        weight : np.array
            The weight of each sample.

        Returns
        -------
        pd.DataFrame
        """
        self.fit(X, treatment, weight)
        return self.transform()


class AttributeEffect:
    """Estimating the effect of the intervention by attribute."""

    def __init__(self) -> None:
        self.effect = None

    def fit(
        self,
        X: pd.DataFrame,
        treatment: pd.Series,
        y: pd.Series,
        weight: Union[np.ndarray, None] = None,
    ) -> None:
        """Fit the model with X, y and weight.

        Parameters
        ----------
        X : pd.DataFrame
            Covariates for propensity score.
        treatment : pd.Series
            Flags with or without intervention.
        y : pd.Series
            Outcome variables.
        weight : np.array
            The weight of each sample.

        Returns
        -------
        None
        """
        if weight is None:
            weight = np.ones(X.shape[0])

        is_treat = treatment == 1

        self.treat_result = sm.WLS(
            y[is_treat],
            X[is_treat],
            weights=weight[is_treat],  # type: ignore
        ).fit()
        self.control_result = sm.WLS(
            y[~is_treat],
            X[~is_treat],
            weights=weight[~is_treat],  # type: ignore
        ).fit()

    def transform(self):
        """Apply the estimating the effect of the intervention by attribute.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame
        """
        result = pd.DataFrame()
        models = [self.control_result, self.treat_result]
        for i, model in enumerate(models):
            result[f"Z{i}_effect"] = model.params.round(1)
            result[f"Z{i}_tvalue"] = model.tvalues.round(2).apply(
                lambda x: str(x) + "**" if abs(x) >= 1.96 else str(x)
            )

        # Estimate Lift Values
        result["Lift"] = result["Z1_effect"] - result["Z0_effect"]
        result_df = result.sort_values(by="Lift")
        self.effect = result_df
        return result_df

    def plot_lift_values(self, figsize: Tuple[float, float] = (12, 6)) -> None:
        # TODO: Move to visualize.py
        """Plot the effect.

        Parameters
        ----------
        figsize : tuple
            Figure dimension ``(width, height)`` in inches.

        Returns
        -------
        """
        plt.figure(figsize=figsize)
        plt.title("Treatment Lift Values")
        plt.bar(self.effect.index, self.effect["Lift"].values)  # type: ignore
        plt.ylabel("Lift Value")
        plt.xticks(rotation=90)
        plt.tight_layout()
        plt.show()


class VIF:
    """Variance Inflation Factor (VIF)."""

    def __init__(self) -> None:
        self.result = None

    def fit(self, data: pd.DataFrame) -> None:
        """Fit the model with data.

        Parameters
        ----------
        data : pd.DataFrame

        Returns
        -------
        None
        """
        vif = pd.DataFrame(
            index=data.columns.tolist(), columns=["VIF"], dtype="float64"
        )

        for feature in data.columns.tolist():
            X = data.drop([feature], axis=1)
            y = data[feature]

            model = linear_model.OLS(endog=y, exog=X)
            r2 = model.fit().rsquared
            vif.loc[feature, "VIF"] = np.round(1 / (1 - r2), 2)
        self.result = vif

    def transform(self):
        """Apply the calculating vif.

        Returns
        -------
        result : pd.DataFrame
        """
        return self.result

    def fit_transform(self, data: pd.DataFrame, **kwargs):
        """Fit the model with data and apply the calculating vif.

        Parameters
        ----------
        data : pd.DataFrame

        Returns
        -------
        result : pd.DataFrame
        """
        self.fit(data, **kwargs)
        return self.transform()


def f1_score(
    y_true: np.ndarray,
    y_score: np.ndarray,
    threshold: float = 0.5,
    is_auto: bool = True,
):
    """Calculate the F1 score.

    Parameters
    ----------
    y_true : numpy.ndarray
        The target vector.
    y_score : numpy.ndarray
        The score vector.
    threshold : 'auto' or float
        Increasing thresholds on the decision function
        used to compute precision and recall.

    Returns
    -------
    score : float
    """
    assert 0 <= threshold < 1, "mode must be or 0 to 1."

    if is_auto:
        fpr, tpr, thresholds = metrics.roc_curve(y_true, y_score)
        gmeans = np.sqrt(tpr * (1 - fpr))
        threshold = thresholds[np.argmax(gmeans)]

    score = metrics.f1_score(y_true, (y_score > threshold))
    return score



from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from sklearn import metrics

from .metrics import EffectSize


def plot_effect_size(
    X: np.ndarray,
    treatment: np.ndarray,
    weight: np.ndarray | None = None,
    ascending: bool = False,
    sortbyraw: bool = True,
    figsize: Tuple[float, float] = (12, 6),
    threshold: float = 0.2,
) -> None:
    """
    Plot the effects of the intervention.

    Parameters
    ----------
    X : numpy.ndarray
        Covariates for propensity score.
    treatment : numpy.ndarray
        Flags with or without intervention.
    weight : numpy.ndarray
        The weight of each sample
    ascending : bool
        Sort in ascending order.
    sortbyraw : bool
        Flags with sort by raw data or weighted data.
    figsize : tuple
        Figure dimension ``(width, height)`` in inches.
    threshold : float
        Threshold value for effect size.

    Returns
    -------
    None
    """
    es = EffectSize()
    es.fit(X, treatment, weight=weight)
    ajusted_names, ajusted_effects = es.transform()

    es = EffectSize()
    es.fit(X, treatment, weight=None)
    raw_names, raw_effects = es.transform()

    sort_data = raw_effects if sortbyraw else ajusted_effects

    if ascending:
        sorted_index = np.argsort(sort_data)
    else:
        sorted_index = np.argsort(sort_data)[::-1]

    plt.figure(figsize=figsize)
    plt.title("Standard Diff")

    plt.bar(
        raw_names[sorted_index],
        raw_effects[sorted_index],
        color="tab:blue",
        label="Raw",
    )
    plt.bar(
        ajusted_names[sorted_index],
        ajusted_effects[sorted_index],
        color="tab:cyan",
        label="Ajusted",
        width=0.5,
    )
    plt.ylabel("d value")
    plt.xticks(rotation=90)
    plt.plot(
        [0.0, len(raw_names)],
        [threshold, threshold],
        color="tab:red",
        linestyle="--",
    )
    plt.tight_layout()
    plt.legend()
    plt.show()


def plot_roc_curve(y_true, y_score, figsize=(7, 6)) -> None:
    """Plot the roc curve.

    Parameters
    ----------
    y_true : numpy.ndarray
        The target vector.
    y_score : numpy.ndarray
        The score vector.
    figsize : tuple
        Figure dimension ``(width, height)`` in inches.

    Returns
    -------
    None
    """
    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_score)
    auc = metrics.auc(fpr, tpr)

    plt.figure(figsize=figsize)
    plt.plot(
        fpr,
        tpr,
        color="darkorange",
        lw=2,
        label="ROC curve (area = %0.2f)" % auc,
    )
    plt.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--")
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.title("ROC Curve")
    plt.legend(loc="lower right")
    plt.show()


def plot_probability_distribution(y_true, y_score, figsize=(12, 6)) -> None:
    """Plot propensity scores, color-coded by
    the presence or absence of intervention.

    Parameters
    ----------
    y_true : numpy.ndarray
        The target vector.
    y_score : numpy.ndarray
        The score vector.
    figsize : tuple
        Figure dimension ``(width, height)`` in inches.

    Returns
    -------
    None
    """
    bins_list = list(np.linspace(0, 1, 100, endpoint=False))
    plt.figure(figsize=figsize)
    plt.title("Probability Distoribution.")
    plt.xlabel("Probability")
    plt.ylabel("Number of Data")
    plt.hist(
        y_score[y_true == 0],
        bins=bins_list,
        rwidth=0.4,
        align="left",
        color="tab:blue",
    )
    plt.hist(
        y_score[y_true == 1],
        bins=bins_list,
        rwidth=0.4,
        align="mid",
        color="tab:orange",
    )
    plt.show()


def plot_treatment_effect(
    outcome_name,
    control_effect,
    treat_effect,
    effect_size,
    figsize=None,
    fontsize=12,
) -> None:
    """Plot the effects of the intervention.

    Parameters
    ----------
    outcome_name : str
        Outcome name. it use for figure title.
    control_effect : float or int
        Average control Group Effect size.
    treat_effect : float or int
        Average treatment Group Effect size.
    effect_size : float or int
        Treatment Effect size.
    figsize : tuple
        Figure dimension ``(width, height)`` in inches.
    fontsize: int
        The font size of the text. See `.Text.set_size` for possible values.

    Returns
    -------
    None
    """
    plt.figure(figsize=figsize)
    plt.title(outcome_name)
    plt.bar(
        ["control", "treatment"],
        [control_effect, treat_effect],
        label=f"Treatment Effect : {effect_size}",
    )
    plt.ylabel("effect size")
    plt.legend(loc="upper left", fontsize=fontsize)
    plt.show()


def plot_auuc(uplift_score, lift, baseline, auuc=None) -> None:
    """Plot Area Under the Uplift Curve (AUUC).

    Parameters
    ----------
    uplift_score : numpy.ndarray
        Array of uplift scores.
    lift : numpy.ndarray
        Array of lift, treatment effect.
    baseline : numpy.ndarray
        Array of random treat effect.
    auuc : float
        AUUC score.

    Returns
    -------
    None
    """
    label = f"AUUC = {auuc:.4f}" if auuc is not None else None

    plt.title("AUUC")
    plt.plot(lift, label=label)
    plt.plot(baseline)
    plt.xlabel("uplift score rank")
    plt.ylabel("lift")
    plt.legend(loc="lower right")
    plt.show()



import copy

import numpy as np
from sklearn.neighbors import NearestNeighbors


class Matching:
    """
    Matching with propensity score.

    Attributes
    ----------
    p_score : numpy.ndarray
        Propensity Score.
    """

    def __init__(self, learner, min_match_dist=1e-2) -> None:
        """
        Parameters
        ----------
        learner :
            Learner to estimate propensity score.
        """
        self.p_score: np.ndarray
        self.eps: float = 1e-8

        self.learner = learner
        self.min_match_dist = min_match_dist

    def fit(self, X, treatment, y) -> None:
        """
        Fit learner and Estimate Propensity Score.

        Parameters
        ----------
        X : numpy.ndarray
            Covariates for propensity score.
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        esp : float
            Extreme Value Trend Score Rounding Value.
        """
        self.learner.fit(X, treatment)
        self.p_score = np.clip(
            self.learner.predict_proba(X)[:, 1], self.eps, 1 - self.eps
        )

    def get_score(self):
        """
        Return propensity score.
        """
        return self.p_score

    def get_weight(self, treatment, mode="ate"):
        """
        Return sample weight representing matching.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        mode : str
            Adjustment method. raw or ate.

        Returns
        -------
        sampel_weight : numpy.ndarray
        """
        self._check_mode(mode)
        if mode == "raw":
            return np.ones(treatment.shape[0])
        elif mode == "ate":
            return self._get_matching_weight(treatment)
        else:
            raise ValueError("mode must be raw or ate.")

    def _check_mode(self, mode):
        """
        Check if it is a supported mode.

        Parameters
        ----------
        mode : str
            Adjustment method. raw or ate.
        """
        mode_list = ["raw", "ate"]
        assert mode in mode_list, "mode must be string and it is raw　or ate."

    def _get_matching_weight(self, treatment):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.

        Returns
        -------
        sampel_weight : numpy.ndarray
        """
        score = self.p_score
        treat_idx, control_idx = self._get_nearest_idx(treatment, score)

        matching_idx = np.concatenate((treat_idx, control_idx), axis=0)
        idx, counts = np.unique(matching_idx, return_counts=True)
        weight = np.zeros(treatment.shape[0])
        weight[idx] = counts
        return weight

    def _get_nearest_idx(self, treatment, score):
        """
        Match the closest data between groups with and without intervention.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        score : numpy.ndarray
            Propensity Score.

        Returns
        -------
        treat_idx : numpy.ndarray
            Sample index of treatment group.
        control_idx : numpy.ndarray
            Sample index of control group.
        """
        score = score.reshape(-1, 1)
        control_size, treat_size = (~treatment).sum(), treatment.sum()
        major_sample_group = np.argmax([control_size, treat_size])

        neigh = NearestNeighbors(n_neighbors=5, metric="manhattan")
        neigh.fit(score[treatment == major_sample_group])
        distance, match_idx = neigh.kneighbors(
            score[treatment != major_sample_group], 1, return_distance=True
        )
        match_idx = match_idx[distance < self.min_match_dist].flatten()

        if major_sample_group == 1:
            treat_idx = np.where(treatment)[0][match_idx]
            control_idx = np.where(~treatment)[0]
        else:
            treat_idx = np.where(treatment)[0]
            control_idx = np.where(~treatment)[0][match_idx]
        return treat_idx, control_idx

    def estimate_effect(self, treatment, y, mode="ate"):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        mode : str
            Adjustment method. raw or ate.

        Returns
        -------
        ajusted_outcomes : numpy.ndarray
        """
        self._check_mode(mode)
        weight = self.get_weight(treatment, mode=mode)
        return self._estimate_outcomes(treatment, y, weight)

    def _estimate_outcomes(self, treatment, y, weight):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        weight : numpy.ndarray
            sample weight with matching.

        Returns
        -------
        avg_y_control : float
            average outcome of control group.
        avg_y_treat : float
            average outcome of treatment group.
        effect_size : float
            diff of average_y_treatment and average_y_control
        """
        avg_y_control = np.average(
            y[~treatment], axis=0, weights=weight[~treatment]
        )
        avg_y_treat = np.average(
            y[treatment], axis=0, weights=weight[treatment]
        )
        effect_size = avg_y_treat - avg_y_control
        return (avg_y_control, avg_y_treat, effect_size)


class IPW:
    """Inverse Probability Weighting Method."""

    def __init__(self, learner) -> None:
        """
        Parameters
        ----------
        learner :
            Learner to estimate propensity score.
        """
        self.p_score: np.ndarray

        self.learner = learner

    def fit(
        self,
        X: np.ndarray,
        treatment: np.ndarray,
        y: np.ndarray,
        eps: float = 1e-8,
    ) -> None:
        """
        Fit learner and Estimate Propensity Score.

        Parameters
        ----------
        X : numpy.ndarray
            Covariates for propensity score.
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        esp : float
            Extreme Value Trend Score Rounding Value.
        """
        self.learner.fit(X, treatment)
        assert 0 <= eps < 1, "clip must be 0 to 1."
        pred = self.learner.predict_proba(X)[:, 1]
        self.p_score = np.clip(pred, eps, 1 - eps)

    def get_score(self) -> np.ndarray:
        """
        Return propensity score.
        """
        return self.p_score

    def get_weight(self, treatment, mode="ate"):
        """
        Return sample weight representing matching.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        mode : str
            Adjustment method. must be raw, ate, att or atu.

        Returns
        -------
        sampel_weight : numpy.ndarray
        """
        self._check_mode(mode)
        if mode == "raw":
            return np.ones(treatment.shape[0])
        elif mode == "ate":
            return np.where(
                treatment == 1, 1 / self.p_score, 1 / (1 - self.p_score)
            )
        elif mode == "att":
            return np.where(
                treatment == 1, 1, self.p_score / (1 - self.p_score)
            )
        elif mode == "atu":
            return np.where(
                treatment == 1, (1 - self.p_score) / self.p_score, 1
            )
        else:
            raise ValueError("mode must be raw, ate, att or atu.")

    def estimate_effect(self, treatment, y, mode="ate"):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        mode : str
            Adjustment method. must be raw, ate, att or atu.

        Returns
        -------
        ajusted_outcomes : numpy.ndarray
        """
        self._check_mode(mode)
        weight = self.get_weight(treatment, mode=mode)
        return self._estimate_outcomes(treatment, y, weight)

    def _check_mode(self, mode):
        """
        Check if it is a supported mode.

        Parameters
        ----------
        mode : str
            Adjustment method. must be raw, ate, att or atu.
        """
        mode_list = ["raw", "ate", "att", "atu"]
        assert (
            mode in mode_list
        ), "mode must be string and it is must be raw, ate, att or atu."

    def _estimate_outcomes(self, treatment, y, weight):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        weight : numpy.ndarray
            sample weight with ipw.

        Returns
        -------
        avg_y_control : float
            average outcome of control group.
        avg_y_treat : float
            average outcome of treatment group.
        effect_size : float
            diff of average_y_treatment and average_y_control
        """
        avg_y_control = np.average(
            y[~treatment], axis=0, weights=weight[~treatment]
        )
        avg_y_treat = np.average(
            y[treatment], axis=0, weights=weight[treatment]
        )
        effect_size = avg_y_treat - avg_y_control
        return (avg_y_control, avg_y_treat, effect_size)


class DoubleRobust(IPW):
    def __init__(self, learner, second_learner) -> None:
        """
        Parameters
        ----------
        learner :
            Learner to estimate propensity score.
        second_learner :
            Learner to estimate anti-real virtual intervention effect.
        """
        super(DoubleRobust, self).__init__(learner)
        self.treat_learner = copy.deepcopy(second_learner)
        self.control_learner = copy.deepcopy(second_learner)

    def fit(
        self,
        X: np.ndarray,
        treatment: np.ndarray,
        y: np.ndarray,
        eps: float = 1e-8,
    ) -> None:
        """
        Fit learner and Estimate Propensity Score.

        Parameters
        ----------
        X : numpy.ndarray
            Covariates for propensity score.
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        esp : float
            Extreme Value Trend Score Rounding Value.
        """
        self.learner.fit(X, treatment)
        assert 0 <= eps < 1, "clip must be 0 to 1."
        self.p_score = np.clip(
            self.learner.predict_proba(X)[:, 1], eps, 1 - eps
        )

        self.y_control = np.zeros(y.shape)
        self.y_treat = np.zeros(y.shape)
        # Fit second models
        for i, _y in enumerate(y.T):
            self.treat_learner.fit(X[treatment], _y[treatment])
            self.control_learner.fit(X[~treatment], _y[~treatment])

            self.y_control[:, i] = np.where(
                ~treatment, _y, self.control_learner.predict(X)
            )
            self.y_treat[:, i] = np.where(
                treatment, _y, self.treat_learner.predict(X)
            )

    def estimate_effect(self, treatment, mode="ate"):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        X : numpy.ndarray
            Covariates for propensity score.
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        mode : str
            Adjustment method. must be raw, ate, att or atu.

        Returns
        -------
        ajusted_outcomes : numpy.ndarray
        """
        self._check_mode(mode)
        weight = self.get_weight(treatment, mode=mode)
        return self._estimate_outcomes(weight)

    def _estimate_outcomes(self, weight):
        """
        Match using propensity score and return sample_weight.

        Parameters
        ----------
        X : numpy.ndarray
            Covariates for propensity score.
        treatment : numpy.ndarray[bool]
            Flags with or without intervention.
        y : numpy.ndarray
            Outcome variables.
        weight : numpy.ndarray
            sample weight of ipw.

        Returns
        -------
        avg_y_control : float
            average outcome of control group.
        avg_y_treat : float
            average outcome of treatment group.
        effect_size : float
            diff of average_y_treatment and average_y_control
        """
        avg_y_control = np.average(self.y_control, axis=0, weights=weight)
        avg_y_treat = np.average(self.y_treat, axis=0, weights=weight)
        effect_size = avg_y_treat - avg_y_control
        return (avg_y_control, avg_y_treat, effect_size)
